class Log(str):
    """A container for logs."""

    def _widget(self):
        from ipywidgets import HTML
        return HTML(
            value="<pre><code>{logs}</code></pre>".format(logs=self)
        )

    def _ipython_display_(self, **kwargs):
        return self._widget()._ipython_display_(**kwargs)


class Logs(dict):
    """A container for multiple logs."""

    def _widget(self):
        from ipywidgets import Accordion
        accordion = Accordion(children=[log._widget() for log in self.values()])
        [accordion.set_title(i, title) for i, title in enumerate(self.keys())]
        return accordion

    def _ipython_display_(self, **kwargs):
        return self._widget()._ipython_display_(**kwargs)







import base64
import getpass
import os
from time import sleep, time
import uuid
import yaml

import dask
import pytest
from dask_kubernetes import KubeCluster, make_pod_spec, ClusterAuth, KubeConfig, KubeAuth
from dask_kubernetes.objects import clean_pod_template
from dask.distributed import Client, wait
from distributed.utils_test import loop, captured_logger  # noqa: F401
from distributed.utils import tmpfile
import kubernetes
from random import random

TEST_DIR = os.path.abspath(os.path.join(__file__, '..'))
CONFIG_DEMO = os.path.join(TEST_DIR, 'config-demo.yaml')
FAKE_CERT = os.path.join(TEST_DIR, 'fake-cert-file')
FAKE_KEY = os.path.join(TEST_DIR, 'fake-key-file')
FAKE_CA = os.path.join(TEST_DIR, 'fake-ca-file')


@pytest.fixture
def api():
    ClusterAuth.load_first()
    return kubernetes.client.CoreV1Api()


@pytest.fixture
def ns(api):
    name = 'test-dask-kubernetes' + str(uuid.uuid4())[:10]
    ns = kubernetes.client.V1Namespace(metadata=kubernetes.client.V1ObjectMeta(name=name))
    api.create_namespace(ns)
    try:
        yield name
    finally:
        api.delete_namespace(name)


@pytest.fixture
def pod_spec(image_name):
    yield make_pod_spec(
        image=image_name,
        extra_container_config={'imagePullPolicy': 'IfNotPresent'}
    )

@pytest.fixture
def clean_pod_spec(pod_spec):
    yield clean_pod_template(pod_spec)


@pytest.fixture
def cluster(pod_spec, ns, loop):
    with KubeCluster(pod_spec, loop=loop, namespace=ns) as cluster:
        yield cluster


@pytest.fixture
def client(cluster):
    with Client(cluster) as client:
        yield client


def test_versions(client):
    client.get_versions(check=True)


def test_basic(cluster, client):
    cluster.scale(2)
    future = client.submit(lambda x: x + 1, 10)
    result = future.result()
    assert result == 11

    while len(cluster.scheduler.workers) < 2:
        sleep(0.1)

    # Ensure that inter-worker communication works well
    futures = client.map(lambda x: x + 1, range(10))
    total = client.submit(sum, futures)
    assert total.result() == sum(map(lambda x: x + 1, range(10)))
    assert all(client.has_what().values())


def test_logs(cluster):
    cluster.scale(2)

    start = time()
    while len(cluster.scheduler.workers) < 2:
        sleep(0.1)
        assert time() < start + 20

    a, b = cluster.pods()
    logs = cluster.logs(a)
    assert 'distributed.worker' in logs

    logs = cluster.logs()
    assert len(logs) == 2
    for pod in logs:
        assert 'distributed.worker' in logs[pod]


def test_ipython_display(cluster):
    ipywidgets = pytest.importorskip('ipywidgets')
    cluster.scale(1)
    cluster._ipython_display_()
    box = cluster._cached_widget
    assert isinstance(box, ipywidgets.Widget)
    cluster._ipython_display_()
    assert cluster._cached_widget is box

    start = time()
    while "<td>1</td>" not in str(box):  # one worker in a table
        assert time() < start + 10
        sleep(0.5)


def test_dask_worker_name_env_variable(pod_spec, loop, ns):
    with dask.config.set({'kubernetes.name': 'foo-{USER}-{uuid}'}):
        with KubeCluster(pod_spec, loop=loop, namespace=ns) as cluster:
            assert 'foo-' + getpass.getuser() in cluster.name


def test_diagnostics_link_env_variable(pod_spec, loop, ns):
    pytest.importorskip('bokeh')
    with dask.config.set({'distributed.dashboard.link': 'foo-{USER}-{port}'}):
        with KubeCluster(pod_spec, loop=loop, namespace=ns) as cluster:
            port = cluster.scheduler.services['bokeh'].port
            cluster._ipython_display_()
            box = cluster._cached_widget

            assert 'foo-' + getpass.getuser() + '-' + str(port) in str(box)


def test_namespace(pod_spec, loop, ns):
    with KubeCluster(pod_spec, loop=loop, namespace=ns) as cluster:
        assert 'dask' in cluster.name
        assert getpass.getuser() in cluster.name
        with KubeCluster(pod_spec, loop=loop, namespace=ns) as cluster2:
            assert cluster.name != cluster2.name

            cluster2.scale(1)
            [pod] = cluster2.pods()


def test_adapt(cluster):
    cluster.adapt()
    with Client(cluster) as client:
        future = client.submit(lambda x: x + 1, 10)
        result = future.result()
        assert result == 11

    start = time()
    while cluster.scheduler.workers:
        sleep(0.1)
        assert time() < start + 10


def test_env(pod_spec, loop, ns):
    with KubeCluster(pod_spec, env={'ABC': 'DEF'}, loop=loop, namespace=ns) as cluster:
        cluster.scale(1)
        with Client(cluster) as client:
            while not cluster.scheduler.workers:
                sleep(0.1)
            env = client.run(lambda: dict(os.environ))
            assert all(v['ABC'] == 'DEF' for v in env.values())


def test_pod_from_yaml(image_name, loop, ns):
    test_yaml = {
        "kind": "Pod",
        "metadata": {
            "labels": {
                "app": "dask",
                "component": "dask-worker"
            }
        },
        "spec": {
            "containers": [{
                "args": [
                    "dask-worker",
                    "$(DASK_SCHEDULER_ADDRESS)",
                    "--nthreads",
                    "1"
                ],
                "image": image_name,
                'imagePullPolicy': 'IfNotPresent',
                "name": "dask-worker"
            }]
        }
    }

    with tmpfile(extension='yaml') as fn:
        with open(fn, mode='w') as f:
            yaml.dump(test_yaml, f)
        with KubeCluster.from_yaml(f.name, loop=loop, namespace=ns) as cluster:
            assert cluster.namespace == ns
            cluster.scale(2)
            with Client(cluster) as client:
                future = client.submit(lambda x: x + 1, 10)
                result = future.result(timeout=10)
                assert result == 11

                start = time()
                while len(cluster.scheduler.workers) < 2:
                    sleep(0.1)
                    assert time() < start + 10, 'timeout'

                # Ensure that inter-worker communication works well
                futures = client.map(lambda x: x + 1, range(10))
                total = client.submit(sum, futures)
                assert total.result() == sum(map(lambda x: x + 1, range(10)))
                assert all(client.has_what().values())


def test_pod_from_yaml_expand_env_vars(image_name, loop, ns):
    try:
        os.environ["FOO_IMAGE"] = image_name

        test_yaml = {
            "kind": "Pod",
            "metadata": {
                "labels": {
                    "app": "dask",
                    "component": "dask-worker"
                }
            },
            "spec": {
                "containers": [{
                    "args": [
                        "dask-worker",
                        "$(DASK_SCHEDULER_ADDRESS)",
                        "--nthreads",
                        "1"
                    ],
                    "image": '${FOO_IMAGE}',
                    'imagePullPolicy': 'IfNotPresent',
                    "name": "dask-worker"
                }]
            }
        }

        with tmpfile(extension='yaml') as fn:
            with open(fn, mode='w') as f:
                yaml.dump(test_yaml, f)
            with KubeCluster.from_yaml(f.name, loop=loop, namespace=ns) as cluster:
                assert cluster.pod_template.spec.containers[0].image == image_name
    finally:
        del os.environ['FOO_IMAGE']


def test_pod_from_dict(image_name, loop, ns):
    spec = {
        'metadata': {},
        'restartPolicy': 'Never',
        'spec': {
            'containers': [{
                'args': ['dask-worker', '$(DASK_SCHEDULER_ADDRESS)',
                         '--nthreads', '1',
                         '--death-timeout', '60'],
                'command': None,
                'image': image_name,
                'imagePullPolicy': 'IfNotPresent',
                'name': 'dask-worker',
            }]
        }
    }

    with KubeCluster.from_dict(spec, loop=loop, namespace=ns) as cluster:
        cluster.scale(2)
        with Client(cluster) as client:
            future = client.submit(lambda x: x + 1, 10)
            result = future.result()
            assert result == 11

            while len(cluster.scheduler.workers) < 2:
                sleep(0.1)

            # Ensure that inter-worker communication works well
            futures = client.map(lambda x: x + 1, range(10))
            total = client.submit(sum, futures)
            assert total.result() == sum(map(lambda x: x + 1, range(10)))
            assert all(client.has_what().values())


def test_pod_from_minimal_dict(image_name, loop, ns):
    spec = {
        'spec': {
            'containers': [{
                'args': ['dask-worker', '$(DASK_SCHEDULER_ADDRESS)',
                         '--nthreads', '1',
                         '--death-timeout', '60'],
                'command': None,
                'image': image_name,
                'imagePullPolicy': 'IfNotPresent',
                'name': 'worker'
            }]
        }
    }

    with KubeCluster.from_dict(spec, loop=loop, namespace=ns) as cluster:
        cluster.adapt()
        with Client(cluster) as client:
            future = client.submit(lambda x: x + 1, 10)
            result = future.result()
            assert result == 11


def test_pod_template_from_conf():
    spec = {'spec': {'containers': [{'name': 'some-name'}]}}

    with dask.config.set({'kubernetes.worker-template': spec}):
        with KubeCluster() as cluster:
            assert cluster.pod_template.spec.containers[0].name == 'some-name'


def test_bad_args(loop):
    with pytest.raises(TypeError) as info:
        KubeCluster('myfile.yaml')

    assert 'KubeCluster.from_yaml' in str(info.value)

    with pytest.raises((ValueError, TypeError)) as info:
        KubeCluster({'kind': 'Pod'})

    assert 'KubeCluster.from_dict' in str(info.value)


def test_constructor_parameters(pod_spec, loop, ns):
    env = {'FOO': 'BAR', 'A': 1}
    with KubeCluster(pod_spec, name='myname', namespace=ns, loop=loop, env=env) as cluster:
        pod = cluster.pod_template
        assert pod.metadata.namespace == ns

        var = [v for v in pod.spec.containers[0].env if v.name == 'FOO']
        assert var and var[0].value == 'BAR'

        var = [v for v in pod.spec.containers[0].env if v.name == 'A']
        assert var and var[0].value == '1'

        assert pod.metadata.generate_name == 'myname'


def test_reject_evicted_workers(cluster):
    cluster.scale(1)

    start = time()
    while len(cluster.scheduler.workers) != 1:
        sleep(0.1)
        assert time() < start + 60

    # Evict worker
    [worker] = cluster.pods()
    cluster.core_api.create_namespaced_pod_eviction(
        worker.metadata.name,
        worker.metadata.namespace,
        kubernetes.client.V1beta1Eviction(
            delete_options=kubernetes.client.V1DeleteOptions(grace_period_seconds=300),
            metadata=worker.metadata))

    # Wait until pod is evicted
    start = time()
    while cluster.pods()[0].status.phase == 'Running':
        sleep(0.1)
        assert time() < start + 60

    [worker] = cluster.pods()
    assert worker.status.phase == 'Failed'

    # Make sure the failed pod is removed
    pods = cluster._cleanup_terminated_pods([worker])
    assert len(pods) == 0

    start = time()
    while cluster.pods():
        sleep(0.1)
        assert time() < start + 60


def test_scale_up_down(cluster, client):
    np = pytest.importorskip('numpy')
    cluster.scale(2)

    start = time()
    while len(cluster.scheduler.workers) != 2:
        sleep(0.1)
        assert time() < start + 10

    a, b = list(cluster.scheduler.workers)
    x = client.submit(np.ones, 1, workers=a)
    y = client.submit(np.ones, 50_000_000, workers=b)

    wait([x, y])

    start = time()
    while (cluster.scheduler.workers[a].metrics['memory'] >
           cluster.scheduler.workers[b].metrics['memory']):
        sleep(0.1)
        assert time() < start + 1

    cluster.scale(1)

    start = time()
    while len(cluster.scheduler.workers) != 1:
        sleep(0.1)
        assert time() < start + 10

    assert set(cluster.scheduler.workers) == {b}


def test_scale_up_down_fast(cluster, client):
    cluster.scale(1)

    start = time()
    while len(cluster.scheduler.workers) != 1:
        sleep(0.1)
        assert time() < start + 10

    worker = next(iter(cluster.scheduler.workers.values()))

    # Put some data on this worker
    future = client.submit(lambda: b'\x00' * int(1e6))
    wait(future)
    assert worker in cluster.scheduler.tasks[future.key].who_has

    # Rescale the cluster many times without waiting: this should put some
    # pressure on kubernetes but this should never fail nor delete our worker
    # with the temporary result.
    for i in range(10):
        cluster.scale(4)
        sleep(random() / 2)
        cluster.scale(1)
        sleep(random() / 2)

    start = time()
    while len(cluster.scheduler.workers) != 1:
        sleep(0.1)
        assert time() < start + 10

    # The original task result is still stored on the original worker: this pod
    # has never been deleted when rescaling the cluster and the result can
    # still be fetched back.
    assert worker in cluster.scheduler.tasks[future.key].who_has
    assert len(future.result()) == int(1e6)


def test_scale_down_pending(cluster, client):
    # Try to scale the cluster to use more pods than available
    nodes = cluster.core_api.list_node().items
    max_pods = sum(int(node.status.allocatable['pods']) for node in nodes)
    if max_pods > 50:
        # It's probably not reasonable to run this test against a large
        # kubernetes cluster.
        pytest.skip("Require a small test kubernetes cluster (maxpod <= 50)")
    extra_pods = 5
    requested_pods = max_pods + extra_pods
    cluster.scale(requested_pods)

    start = time()
    while len(cluster.scheduler.workers) < 2:
        sleep(0.1)
        # Wait a bit because the kubernetes cluster can take time to provision
        # the requested pods as we requested a large number of pods.
        assert time() < start + 60

    pending_pods = [p for p in cluster.pods() if p.status.phase == 'Pending']
    assert len(pending_pods) >= extra_pods

    running_workers = list(cluster.scheduler.workers.keys())
    assert len(running_workers) >= 2

    # Put some data on those workers to make them important to keep as long
    # as possible.
    def load_data(i):
        return b'\x00' * (i * int(1e6))

    futures = [client.submit(load_data, i, workers=w)
               for i, w in enumerate(running_workers)]
    wait(futures)

    # Reduce the cluster size down to the actually useful nodes: pending pods
    # and running pods without results should be shutdown and removed first:
    cluster.scale(len(running_workers))

    start = time()
    pod_statuses = [p.status.phase for p in cluster.pods()]
    while len(pod_statuses) != len(running_workers):
        if time() - start > 60:
            raise AssertionError("Expected %d running pods but got %r"
                                 % (len(running_workers), pod_statuses))
        sleep(0.1)
        pod_statuses = [p.status.phase for p in cluster.pods()]

    assert pod_statuses == ['Running'] * len(running_workers)
    assert list(cluster.scheduler.workers.keys()) == running_workers

    # Terminate everything
    cluster.scale(0)

    start = time()
    while len(cluster.scheduler.workers) > 0:
        sleep(0.1)
        assert time() < start + 60


def test_automatic_startup(image_name, loop, ns):
    test_yaml = {
        "kind": "Pod",
        "metadata": {
            "labels": {
                "foo": "bar",
            }
        },
        "spec": {
            "containers": [{
                "args": [
                    "dask-worker",
                    "$(DASK_SCHEDULER_ADDRESS)",
                    "--nthreads",
                    "1"
                ],
                "image": image_name,
                "name": "dask-worker"
            }]
        }
    }

    with tmpfile(extension='yaml') as fn:
        with open(fn, mode='w') as f:
            yaml.dump(test_yaml, f)
        with dask.config.set({'kubernetes.worker-template-path': fn}):
            with KubeCluster(loop=loop, namespace=ns) as cluster:
                assert cluster.pod_template.metadata.labels['foo'] == 'bar'


def test_repr(cluster):
    for text in [repr(cluster), str(cluster)]:
        assert 'Box' not in text
        assert cluster.scheduler.address in text
        assert "workers=0" in text


def test_escape_username(pod_spec, loop, ns, monkeypatch):
    monkeypatch.setenv('LOGNAME', 'foo!')

    with KubeCluster(pod_spec, loop=loop, namespace=ns) as cluster:
        assert 'foo' in cluster.name
        assert '!' not in cluster.name
        assert 'foo' in cluster.pod_template.metadata.labels['user']


def test_escape_name(pod_spec, loop, ns):
    with KubeCluster(pod_spec, loop=loop, namespace=ns, name='foo@bar') as cluster:
        assert '@' not in str(cluster.pod_template)


def test_maximum(cluster):
    with dask.config.set(**{'kubernetes.count.max': 1}):
        with captured_logger('dask_kubernetes') as logger:
            cluster.scale(10)

            start = time()
            while len(cluster.scheduler.workers) <= 0:
                sleep(0.1)
                assert time() < start + 60

            sleep(0.5)
            assert len(cluster.scheduler.workers) == 1

        result = logger.getvalue()
        assert "scale beyond maximum number of workers" in result.lower()


def test_default_toleration(clean_pod_spec):
    tolerations = clean_pod_spec.to_dict()['spec']['tolerations']
    assert {
        'key': 'k8s.dask.org/dedicated',
        'operator': 'Equal',
        'value': 'worker',
        'effect': 'NoSchedule',
        'toleration_seconds': None
    } in tolerations
    assert {
        'key': 'k8s.dask.org_dedicated',
        'operator': 'Equal',
        'value': 'worker',
        'effect': 'NoSchedule',
        'toleration_seconds': None
    } in tolerations


def test_default_toleration_preserved(image_name):
    pod_spec = make_pod_spec(
        image=image_name,
        extra_pod_config={
            'tolerations': [
                {
                    'key': 'example.org/toleration',
                    'operator': 'Exists',
                    'effect': 'NoSchedule',
                }
            ],
        }
    )
    cluster = KubeCluster(pod_spec)
    tolerations = cluster.pod_template.to_dict()['spec']['tolerations']
    assert {
        'key': 'k8s.dask.org/dedicated',
        'operator': 'Equal',
        'value': 'worker',
        'effect': 'NoSchedule',
        'toleration_seconds': None
    } in tolerations
    assert {
        'key': 'k8s.dask.org_dedicated',
        'operator': 'Equal',
        'value': 'worker',
        'effect': 'NoSchedule',
        'toleration_seconds': None
    } in tolerations
    assert {
        'key': 'example.org/toleration',
        'operator': 'Exists',
        'effect': 'NoSchedule',
    } in tolerations


def test_default_affinity(clean_pod_spec):
    affinity = clean_pod_spec.to_dict()['spec']['affinity']

    assert {
        'key': 'k8s.dask.org/node-purpose',
        'operator': 'In',
        'values': ['worker']
    } in affinity['node_affinity']['preferred_during_scheduling_ignored_during_execution'][0]['preference']['match_expressions']
    assert affinity['node_affinity']['preferred_during_scheduling_ignored_during_execution'][0]['weight'] == 100
    assert affinity['node_affinity']['required_during_scheduling_ignored_during_execution'] is None
    assert affinity['pod_affinity'] is None


def test_auth_missing(pod_spec, ns, loop):
    with pytest.raises(kubernetes.config.ConfigException) as info:
        KubeCluster(pod_spec, auth=[], loop=loop, namespace=ns)

    assert "No authorization methods were provided" in str(info.value)


def test_auth_tries_all_methods(pod_spec, ns, loop):
    fails = {'count': 0}
    class FailAuth(ClusterAuth):
        def load(self):
            fails['count'] += 1
            raise kubernetes.config.ConfigException('Fail #{count}'.format(**fails))

    with pytest.raises(kubernetes.config.ConfigException) as info:
        KubeCluster(pod_spec, auth=[FailAuth()] * 3, loop=loop, namespace=ns)

    assert "Fail #3" in str(info.value)
    assert fails['count'] == 3


def test_auth_kubeconfig_with_filename():
    KubeConfig(config_file=CONFIG_DEMO).load()

    # we've set the default configuration, so check that it is default
    config = kubernetes.client.Configuration()
    assert config.host == 'https://1.2.3.4'
    assert config.cert_file == FAKE_CERT
    assert config.key_file == FAKE_KEY
    assert config.ssl_ca_cert == FAKE_CA


def test_auth_kubeconfig_with_context():
    KubeConfig(config_file=CONFIG_DEMO, context='exp-scratch').load()

    # we've set the default configuration, so check that it is default
    config = kubernetes.client.Configuration()
    assert config.host == 'https://5.6.7.8'
    assert config.api_key['authorization'] == 'Basic {}'.format(
        base64.b64encode(b'exp:some-password').decode('ascii')
    )


def test_auth_explicit():
    KubeAuth(host='https://9.8.7.6', username='abc', password='some-password').load()

    config = kubernetes.client.Configuration()
    assert config.host == 'https://9.8.7.6'
    assert config.username == 'abc'
    assert config.password == 'some-password'
    assert config.get_basic_auth_token() == 'Basic {}'.format(
        base64.b64encode(b'abc:some-password').decode('ascii')
    )



import pytest


def pytest_addoption(parser):
    parser.addoption(
        "--worker-image",
        help="Worker image to use for testing"
    )


@pytest.fixture
def image_name(request):
    worker_image = request.config.getoption('--worker-image')
    if not worker_image:
        pytest.fail("Need to pass --worker-image. "
                    "Image must have the same Python and dask versions as host")
    return worker_image



import getpass
import logging
import os
import socket
import string
import time
from urllib.parse import urlparse
import uuid
from weakref import finalize

try:
    import yaml
except ImportError:
    yaml = False

import dask
from distributed.deploy import LocalCluster, Cluster
from distributed.comm.utils import offload
import kubernetes
from tornado import gen

from .objects import make_pod_from_dict, clean_pod_template
from .auth import ClusterAuth
from .logs import Log, Logs

logger = logging.getLogger(__name__)


class KubeCluster(Cluster):
    """ Launch a Dask cluster on Kubernetes

    This starts a local Dask scheduler and then dynamically launches
    Dask workers on a Kubernetes cluster. The Kubernetes cluster is taken
    to be either the current one on which this code is running, or as a
    fallback, the default one configured in a kubeconfig file.

    **Environments**

    Your worker pod image should have a similar environment to your local
    environment, including versions of Python, dask, cloudpickle, and any
    libraries that you may wish to use (like NumPy, Pandas, or Scikit-Learn).
    See examples below for suggestions on how to manage and check for this.

    **Network**

    Since the Dask scheduler is launched locally, for it to work, we need to
    be able to open network connections between this local node and all the
    workers nodes on the Kubernetes cluster. If the current process is not
    already on a Kubernetes node, some network configuration will likely be
    required to make this work.

    **Resources**

    Your Kubernetes resource limits and requests should match the
    ``--memory-limit`` and ``--nthreads`` parameters given to the
    ``dask-worker`` command.

    Parameters
    ----------
    pod_template: kubernetes.client.V1Pod
        A Kubernetes specification for a Pod for a dask worker.
    name: str (optional)
        Name given to the pods.  Defaults to ``dask-$USER-random``
    namespace: str (optional)
        Namespace in which to launch the workers.
        Defaults to current namespace if available or "default"
    n_workers: int
        Number of workers on initial launch.
        Use ``scale_up`` to increase this number in the future
    env: Dict[str, str]
        Dictionary of environment variables to pass to worker pod
    host: str
        Listen address for local scheduler.  Defaults to 0.0.0.0
    port: int
        Port of local scheduler
    auth: List[ClusterAuth] (optional)
        Configuration methods to attempt in order.  Defaults to
        ``[InCluster(), KubeConfig()]``.
    **kwargs: dict
        Additional keyword arguments to pass to LocalCluster

    Examples
    --------
    >>> from dask_kubernetes import KubeCluster, make_pod_spec
    >>> pod_spec = make_pod_spec(image='daskdev/dask:latest',
    ...                          memory_limit='4G', memory_request='4G',
    ...                          cpu_limit=1, cpu_request=1,
    ...                          env={'EXTRA_PIP_PACKAGES': 'fastparquet git+https://github.com/dask/distributed'})
    >>> cluster = KubeCluster(pod_spec)
    >>> cluster.scale(10)

    You can also create clusters with worker pod specifications as dictionaries
    or stored in YAML files

    >>> cluster = KubeCluster.from_yaml('worker-template.yml')
    >>> cluster = KubeCluster.from_dict({...})

    Rather than explicitly setting a number of workers you can also ask the
    cluster to allocate workers dynamically based on current workload

    >>> cluster.adapt()

    You can pass this cluster directly to a Dask client

    >>> from dask.distributed import Client
    >>> client = Client(cluster)

    You can verify that your local environment matches your worker environments
    by calling ``client.get_versions(check=True)``.  This will raise an
    informative error if versions do not match.

    >>> client.get_versions(check=True)

    The ``daskdev/dask`` docker images support ``EXTRA_PIP_PACKAGES``,
    ``EXTRA_APT_PACKAGES`` and ``EXTRA_CONDA_PACKAGES`` environment variables
    to help with small adjustments to the worker environments.  We recommend
    the use of pip over conda in this case due to a much shorter startup time.
    These environment variables can be modified directly from the KubeCluster
    constructor methods using the ``env=`` keyword.  You may list as many
    packages as you like in a single string like the following:

    >>> pip = 'pyarrow gcsfs git+https://github.com/dask/distributed'
    >>> conda = '-c conda-forge scikit-learn'
    >>> KubeCluster.from_yaml(..., env={'EXTRA_PIP_PACKAGES': pip,
    ...                                 'EXTRA_CONDA_PACKAGES': conda})

    You can also start a KubeCluster with no arguments *if* the worker template
    is specified in the Dask config files, either as a full template in
    ``kubernetes.worker-template`` or a path to a YAML file in
    ``kubernetes.worker-template-path``.

    See https://docs.dask.org/en/latest/configuration.html for more
    information about setting configuration values.::

        $ export DASK_KUBERNETES__WORKER_TEMPLATE_PATH=worker_template.yaml

    >>> cluster = KubeCluster()  # automatically finds 'worker_template.yaml'

    See Also
    --------
    KubeCluster.from_yaml
    KubeCluster.from_dict
    KubeCluster.adapt
    """
    def __init__(
            self,
            pod_template=None,
            name=None,
            namespace=None,
            n_workers=None,
            host=None,
            port=None,
            env=None,
            auth=ClusterAuth.DEFAULT,
            **kwargs
    ):
        name = name or dask.config.get('kubernetes.name')
        namespace = namespace or dask.config.get('kubernetes.namespace')
        n_workers = n_workers if n_workers is not None else dask.config.get('kubernetes.count.start')
        host = host or dask.config.get('kubernetes.host')
        port = port if port is not None else dask.config.get('kubernetes.port')
        env = env if env is not None else dask.config.get('kubernetes.env')

        if not pod_template and dask.config.get('kubernetes.worker-template', None):
            d = dask.config.get('kubernetes.worker-template')
            d = dask.config.expand_environment_variables(d)
            pod_template = make_pod_from_dict(d)

        if not pod_template and dask.config.get('kubernetes.worker-template-path', None):
            import yaml
            fn = dask.config.get('kubernetes.worker-template-path')
            fn = fn.format(**os.environ)
            with open(fn) as f:
                d = yaml.safe_load(f)
            d = dask.config.expand_environment_variables(d)
            pod_template = make_pod_from_dict(d)

        if not pod_template:
            msg = ("Worker pod specification not provided. See KubeCluster "
                   "docstring for ways to specify workers")
            raise ValueError(msg)

        self.cluster = LocalCluster(ip=host or socket.gethostname(),
                                    scheduler_port=port,
                                    n_workers=0, **kwargs)

        ClusterAuth.load_first(auth)

        self.core_api = kubernetes.client.CoreV1Api()

        if namespace is None:
            namespace = _namespace_default()

        name = name.format(user=getpass.getuser(),
                           uuid=str(uuid.uuid4())[:10],
                           **os.environ)
        name = escape(name)

        self.pod_template = clean_pod_template(pod_template)
        # Default labels that can't be overwritten
        self.pod_template.metadata.labels['dask.org/cluster-name'] = name
        self.pod_template.metadata.labels['user'] = escape(getpass.getuser())
        self.pod_template.metadata.labels['app'] = 'dask'
        self.pod_template.metadata.labels['component'] = 'dask-worker'
        self.pod_template.metadata.namespace = namespace

        self.pod_template.spec.containers[0].env.append(
            kubernetes.client.V1EnvVar(name='DASK_SCHEDULER_ADDRESS',
                                       value=self.scheduler_address)
        )
        if env:
            self.pod_template.spec.containers[0].env.extend([
                kubernetes.client.V1EnvVar(name=k, value=str(v))
                for k, v in env.items()
            ])
        self.pod_template.metadata.generate_name = name

        finalize(self, _cleanup_pods, self.namespace, self.pod_template.metadata.labels)

        if n_workers:
            self.scale(n_workers)

    @classmethod
    def from_dict(cls, pod_spec, **kwargs):
        """ Create cluster with worker pod spec defined by Python dictionary

        Examples
        --------
        >>> spec = {
        ...     'metadata': {},
        ...     'spec': {
        ...         'containers': [{
        ...             'args': ['dask-worker', '$(DASK_SCHEDULER_ADDRESS)',
        ...                      '--nthreads', '1',
        ...                      '--death-timeout', '60'],
        ...             'command': None,
        ...             'image': 'daskdev/dask:latest',
        ...             'name': 'dask-worker',
        ...         }],
        ...     'restartPolicy': 'Never',
        ...     }
        ... }
        >>> cluster = KubeCluster.from_dict(spec, namespace='my-ns')  # doctest: +SKIP

        See Also
        --------
        KubeCluster.from_yaml
        """
        return cls(make_pod_from_dict(pod_spec), **kwargs)

    @classmethod
    def from_yaml(cls, yaml_path, **kwargs):
        """ Create cluster with worker pod spec defined by a YAML file

        We can start a cluster with pods defined in an accompanying YAML file
        like the following:

        .. code-block:: yaml

            kind: Pod
            metadata:
              labels:
                foo: bar
                baz: quux
            spec:
              containers:
              - image: daskdev/dask:latest
                name: dask-worker
                args: [dask-worker, $(DASK_SCHEDULER_ADDRESS), --nthreads, '2', --memory-limit, 8GB]
              restartPolicy: Never

        Examples
        --------
        >>> cluster = KubeCluster.from_yaml('pod.yaml', namespace='my-ns')  # doctest: +SKIP

        See Also
        --------
        KubeCluster.from_dict
        """
        if not yaml:
            raise ImportError("PyYaml is required to use yaml functionality, please install it!")
        with open(yaml_path) as f:
            d = yaml.safe_load(f)
            d = dask.config.expand_environment_variables(d)
            return cls.from_dict(d, **kwargs)

    @property
    def namespace(self):
        return self.pod_template.metadata.namespace

    @property
    def name(self):
        return self.pod_template.metadata.generate_name

    def __repr__(self):
        return 'KubeCluster("%s", workers=%d)' % (self.scheduler.address,
                                                  len(self.pods()))

    @property
    def scheduler(self):
        return self.cluster.scheduler

    @property
    def scheduler_address(self):
        return self.scheduler.address

    def pods(self):
        """ A list of kubernetes pods corresponding to current workers

        See Also
        --------
        KubeCluster.logs
        """
        return self.core_api.list_namespaced_pod(
            self.namespace,
            label_selector=format_labels(self.pod_template.metadata.labels)
        ).items

    def logs(self, pod=None):
        """ Logs from a worker pod

        You can get this pod object from the ``pods`` method.

        If no pod is specified all pod logs will be returned. On large clusters
        this could end up being rather large.

        Parameters
        ----------
        pod: kubernetes.client.V1Pod
            The pod from which we want to collect logs.

        See Also
        --------
        KubeCluster.pods
        Client.get_worker_logs
        """
        if pod is None:
            return Logs({pod.status.pod_ip: self.logs(pod) for pod in self.pods()})

        return Log(self.core_api.read_namespaced_pod_log(pod.metadata.name,
                                                         pod.metadata.namespace))

    def scale(self, n):
        """ Scale cluster to n workers

        Parameters
        ----------
        n: int
            Target number of workers

        Example
        -------
        >>> cluster.scale(10)  # scale cluster to ten workers

        See Also
        --------
        KubeCluster.scale_up
        KubeCluster.scale_down
        """
        pods = self._cleanup_terminated_pods(self.pods())
        if n >= len(pods):
            self.scale_up(n, pods=pods)
            return
        else:
            n_to_delete = len(pods) - n
            # Before trying to close running workers, check if we can cancel
            # pending pods (in case the kubernetes cluster was too full to
            # provision those pods in the first place).
            running_workers = list(self.scheduler.workers.keys())
            running_ips = set(urlparse(worker).hostname
                              for worker in running_workers)
            pending_pods = [p for p in pods
                            if p.status.pod_ip not in running_ips]
            if pending_pods:
                pending_to_delete = pending_pods[:n_to_delete]
                logger.debug("Deleting pending pods: %s", pending_to_delete)
                self._delete_pods(pending_to_delete)
                n_to_delete = n_to_delete - len(pending_to_delete)
                if n_to_delete <= 0:
                    return

            to_close = select_workers_to_close(self.scheduler, n_to_delete)
            logger.debug("Closing workers: %s", to_close)
            if len(to_close) < len(self.scheduler.workers):
                # Close workers cleanly to migrate any temporary results to
                # remaining workers.
                @gen.coroutine
                def f(to_close):
                    yield self.scheduler.retire_workers(
                        workers=to_close, remove=True, close_workers=True)
                    yield offload(self.scale_down, to_close)

                self.scheduler.loop.add_callback(f, to_close)
                return

            # Terminate all pods without waiting for clean worker shutdown
            self.scale_down(to_close)

    def _delete_pods(self, to_delete):
        for pod in to_delete:
            try:
                self.core_api.delete_namespaced_pod(
                    pod.metadata.name,
                    self.namespace
                )
                pod_info = pod.metadata.name
                if pod.status.reason:
                    pod_info += ' [{}]'.format(pod.status.reason)
                if pod.status.message:
                    pod_info += ' {}'.format(pod.status.message)
                logger.info('Deleted pod: %s', pod_info)
            except kubernetes.client.rest.ApiException as e:
                # If a pod has already been removed, just ignore the error
                if e.status != 404:
                    raise

    def _cleanup_terminated_pods(self, pods):
        terminated_phases = {'Succeeded', 'Failed'}
        terminated_pods = [p for p in pods if p.status.phase in terminated_phases]
        self._delete_pods(terminated_pods)
        return [p for p in pods if p.status.phase not in terminated_phases]

    def scale_up(self, n, pods=None, **kwargs):
        """
        Make sure we have n dask-workers available for this cluster

        Examples
        --------
        >>> cluster.scale_up(20)  # ask for twenty workers
        """
        maximum = dask.config.get('kubernetes.count.max')
        if maximum is not None and maximum < n:
            logger.info("Tried to scale beyond maximum number of workers %d > %d",
                        n, maximum)
            n = maximum
        pods = pods or self._cleanup_terminated_pods(self.pods())
        to_create = n - len(pods)
        new_pods = []
        for i in range(3):
            try:
                for _ in range(to_create):
                    new_pods.append(self.core_api.create_namespaced_pod(
                        self.namespace, self.pod_template))
                    to_create -= 1
                break
            except kubernetes.client.rest.ApiException as e:
                if e.status == 500 and 'ServerTimeout' in e.body:
                    logger.info("Server timeout, retry #%d", i + 1)
                    time.sleep(1)
                    last_exception = e
                    continue
                else:
                    raise
        else:
            raise last_exception

    def scale_down(self, workers, pods=None):
        """ Remove the pods for the requested list of workers

        When scale_down is called by the _adapt async loop, the workers are
        assumed to have been cleanly closed first and in-memory data has been
        migrated to the remaining workers.

        Note that when the worker process exits, Kubernetes leaves the pods in
        a 'Succeeded' state that we collect here.

        If some workers have not been closed, we just delete the pods with
        matching ip addresses.

        Parameters
        ----------
        workers: List[str] List of addresses of workers to close
        """
        # Get the existing worker pods
        pods = pods or self._cleanup_terminated_pods(self.pods())

        # Work out the list of pods that we are going to delete
        # Each worker to delete is given in the form "tcp://<worker ip>:<port>"
        # Convert this to a set of IPs
        ips = set(urlparse(worker).hostname for worker in workers)
        to_delete = [p for p in pods if p.status.pod_ip in ips]
        if not to_delete:
            return
        self._delete_pods(to_delete)

    def __enter__(self):
        return self

    def close(self, **kwargs):
        """ Close this cluster """
        self.scale_down(self.cluster.scheduler.workers)
        return self.cluster.close(**kwargs)

    def __exit__(self, type, value, traceback):
        _cleanup_pods(self.namespace, self.pod_template.metadata.labels)
        self.cluster.__exit__(type, value, traceback)


def _cleanup_pods(namespace, labels):
    """ Remove all pods with these labels in this namespace """
    api = kubernetes.client.CoreV1Api()
    pods = api.list_namespaced_pod(namespace, label_selector=format_labels(labels))
    for pod in pods.items:
        try:
            api.delete_namespaced_pod(pod.metadata.name, namespace)
            logger.info('Deleted pod: %s', pod.metadata.name)
        except kubernetes.client.rest.ApiException as e:
            # ignore error if pod is already removed
            if e.status != 404:
                raise


def format_labels(labels):
    """ Convert a dictionary of labels into a comma separated string """
    if labels:
        return ','.join(['{}={}'.format(k, v) for k, v in labels.items()])
    else:
        return ''


def _namespace_default():
    """
    Get current namespace if running in a k8s cluster

    If not in a k8s cluster with service accounts enabled, default to
    'default'

    Taken from https://github.com/jupyterhub/kubespawner/blob/master/kubespawner/spawner.py#L125
    """
    ns_path = '/var/run/secrets/kubernetes.io/serviceaccount/namespace'
    if os.path.exists(ns_path):
        with open(ns_path) as f:
            return f.read().strip()
    return 'default'


def select_workers_to_close(scheduler, n_to_close):
    """ Select n workers to close from scheduler """
    workers = list(scheduler.workers.values())
    assert n_to_close <= len(workers)
    key = lambda ws: ws.metrics['memory']
    to_close = set(sorted(scheduler.idle, key=key)[:n_to_close])

    if len(to_close) < n_to_close:
        rest = sorted(workers, key=key, reverse=True)
        while len(to_close) < n_to_close:
            to_close.add(rest.pop())

    return [ws.address for ws in to_close]


valid_characters = string.ascii_letters + string.digits + '_-.'


def escape(s):
    return ''.join(c for c in s if c in valid_characters)



from dask_kubernetes import KubeCluster
from dask_kubernetes.objects import make_pod_spec, make_pod_from_dict
from distributed.utils_test import loop  # noqa: F401


def test_extra_pod_config(image_name, loop):
    """
    Test that our pod config merging process works fine
    """
    cluster = KubeCluster(
        make_pod_spec(
            image_name,
            extra_pod_config={
                'automountServiceAccountToken': False
            }
        ),
        loop=loop,
        n_workers=0,
    )

    pod = cluster.pod_template

    assert pod.spec.automount_service_account_token is False


def test_extra_container_config(image_name, loop):
    """
    Test that our container config merging process works fine
    """
    cluster = KubeCluster(
        make_pod_spec(
            image_name,
            extra_container_config={
                'imagePullPolicy': 'IfNotPresent',
                'securityContext': {
                    'runAsUser': 0
                }
            }
        ),
        loop=loop,
        n_workers=0,
    )

    pod = cluster.pod_template

    assert pod.spec.containers[0].image_pull_policy == 'IfNotPresent'
    assert pod.spec.containers[0].security_context == {
        'runAsUser': 0
    }


def test_container_resources_config(image_name, loop):
    """
    Test container resource requests / limits being set properly
    """
    cluster = KubeCluster(
        make_pod_spec(
            image_name,
            memory_request="1G",
            memory_limit="2G",
            cpu_limit="2",
        ),
        loop=loop,
        n_workers=0,
    )

    pod = cluster.pod_template

    assert pod.spec.containers[0].resources.requests['memory'] == '1G'
    assert pod.spec.containers[0].resources.limits['memory'] == '2G'
    assert pod.spec.containers[0].resources.limits['cpu'] == '2'
    assert "cpu" not in pod.spec.containers[0].resources.requests


def test_extra_container_config_merge(image_name, loop):
    """
    Test that our container config merging process works recursively fine
    """
    cluster = KubeCluster(
        make_pod_spec(
            image_name,
            extra_container_config={
                "env": [ {"name": "BOO", "value": "FOO" } ],
                "args": ["last-item"]
            }
        ),
        loop=loop,
        n_workers=0,
        env={"TEST": "HI"},
    )

    pod = cluster.pod_template

    assert pod.spec.containers[0].env == [
        { "name": "TEST", "value": "HI"},
        { "name": "BOO", "value": "FOO"}
    ]

    assert pod.spec.containers[0].args[-1] == "last-item"


def test_extra_container_config_merge(image_name, loop):
    """
    Test that our container config merging process works recursively fine
    """
    cluster = KubeCluster(
        make_pod_spec(
            image_name,
            env={"TEST": "HI"},
            extra_container_config={
                "env": [ {"name": "BOO", "value": "FOO" } ],
                "args": ["last-item"]
            }
        ),
        loop=loop,
        n_workers=0,
    )

    pod = cluster.pod_template

    for e in [
        { "name": "TEST", "value": "HI"},
        { "name": "BOO", "value": "FOO"}
    ]:
        assert e in pod.spec.containers[0].env

    assert pod.spec.containers[0].args[-1] == "last-item"


def test_make_pod_from_dict():
    d = {
        "kind": "Pod",
        "metadata": {
            "labels": {
                "app": "dask",
                "component": "dask-worker"
            }
        },
        "spec": {
            "containers": [{
                "args": [
                    "dask-worker",
                    "$(DASK_SCHEDULER_ADDRESS)",
                    "--nthreads",
                    "1"
                ],
                "image": "image-name",
                "name": "dask-worker",
                "securityContext": {"capabilities": {"add": ["SYS_ADMIN"]},
                                     "privileged": True},
            }],
            "restartPolicy": "Never",
        }
    }

    pod = make_pod_from_dict(d)

    assert pod.spec.restart_policy == 'Never'
    assert pod.spec.containers[0].security_context.privileged
    assert pod.spec.containers[0].security_context.capabilities.add == ['SYS_ADMIN']



#!/usr/bin/env python

from os.path import exists
from setuptools import setup, find_packages

import versioneer

setup(
    name='dask-kubernetes',
    cmdclass=versioneer.get_cmdclass(),
    version=versioneer.get_version(),
    description='Native Kubernetes integration for Dask',
    url='https://github.com/dask/dask-kubernetes',
    keywords='dask,kubernetes,distributed',
    license='BSD',
    packages=find_packages(),
    include_package_data=True,
    long_description=(open('README.rst').read() if exists('README.rst') else ''),
    zip_safe=False,
    install_requires=list(open('requirements.txt').read().strip().split('\n')),
    python_requires='>=3.5',
)
